﻿using System;
using System.Collections.Generic;
using System.Text;
using System.Collections;
using Editor.Document;
using Editor.Text;
using Editor.Numbering;
using Editor.Collections;

namespace Editor.Document.UseCase
{
    public abstract class TemplateRow : IDocumentRow, ISimpleList<TemplateRow>, IEnumerable<TemplateRow>, IChangeHandler
    {
        protected UseCaseDocument document;
        protected int leftIndent;
        protected int rightIndent;
        protected IText text;
        private EventArgs changeArgs;
        private GapBuffer<TemplateRow> children;
        private bool objectChanging;
        private TemplateRow parent;

        public TemplateRow this[int index]
        {
            get { return children[index]; }
            set { children[index] = value; }
        }

        public bool Changing
        {
            get { return objectChanging; }
        }

        public IDocument Document
        {
            get { return document; }
            internal set { document = value as UseCaseDocument; }
        }

        public abstract bool Editable { get; }
        public abstract bool Formattable { get; }

        public int ItemCount
        {
            get { return children.Length; }
        }

                public int LeftIndent
        {
            get { return leftIndent; }
            set
            {
                if( leftIndent != value )
                {
                    leftIndent = value;

                    IndentChangedEvent( this, EventArgs.Empty );
                }
            }
        }

        public TemplateRow Parent
        {
            get { return parent; }
            internal set { parent = value; }
        }

        public abstract bool Removable { get; }

        public int RightIndent
        {
            get { return rightIndent; }
            set
            {
                if( rightIndent != value )
                {
                    rightIndent = value;

                    IndentChangedEvent( this, EventArgs.Empty );
                }
            }
        }

        public IText Text
        {
            get { return text; }
            set 
            { 
                text.ObjectChanged -= OnTextChanged;
                
                text = value;

                text.ObjectChanged += OnTextChanged;

                ObjectChangedEvent( text, EventArgs.Empty );
            }
        }

        public event EventHandler IndentChanged;
        public event EventHandler ObjectChanging;
        public event EventHandler ObjectChanged;
        public event EventHandler RowAdded;
        public event EventHandler RowRemoved;
        public event EventHandler TextChanged;

        public TemplateRow()
        {
            changeArgs = null;
            children = new GapBuffer<TemplateRow>();
            this.document = null;
            leftIndent = 0;
            objectChanging = false;
            parent = null;
            rightIndent = 0;
            text = new GlyphText();

            text.ObjectChanged += OnTextChanged;           
        }

        public TemplateRow( TemplateRow parent )
        {
            changeArgs = null;
            children = new GapBuffer<TemplateRow>();
            this.document = null;
            leftIndent = 0;
            objectChanging = false;
            this.parent = parent;
            rightIndent = 0;
            text = new GlyphText();

            text.ObjectChanged += OnTextChanged;   
        }

        public void Add( TemplateRow item )
        {
            Insert( children.Length, item );
        }

        public void BeginChange()
        {
            if( !objectChanging )
            {
                objectChanging = true;
                changeArgs = null;

                if( ObjectChanging != null ) ObjectChanging( this, EventArgs.Empty );
            }
        }

        public void Clear()
        {
            Queue<TemplateRow> rowQueue = GetSubRows( 0, children.Length );

            children.SetBuffer( new TemplateRow[0] );

            if( document.Changing )
            {
                while( rowQueue.Count > 0 ) 
                    RowRemovedEvent( rowQueue.Dequeue(), EventArgs.Empty );
            }
            else
            {
                document.BeginChange();

                while( rowQueue.Count > 0 ) 
                    RowRemovedEvent( rowQueue.Dequeue(), EventArgs.Empty );

                document.EndChange( false );
            }
        }

        public abstract TemplateRow CreateNewRow();

        public void EndChange( bool discardEvent )
        {
            if( objectChanging )
            {
                objectChanging = false;

                if( !discardEvent && changeArgs != null ) ObjectChangedEvent( this, changeArgs ); 
            }
        }

        public Delegate[] GetReferences()
        {
            return ObjectChanged.GetInvocationList();
        }

        public IEnumerator<TemplateRow> GetEnumerator()
        {
            yield return this;

            for( int i = 0; i < children.Length; i++ )
            {
                IEnumerator<TemplateRow> enumerator = children[i].GetEnumerator();
                while( enumerator.MoveNext() )
                    yield return enumerator.Current;
            }
        }

        public int IndexOf( TemplateRow item )
        {
            for( int i = 0; i < children.Length; i++ )
                if( object.ReferenceEquals( children[i], item ) ) return i;
            return -1;
        }

        public int IndexOf( TemplateRow item, int startIndex )
        {
            for (int i = startIndex; i < children.Length; i++)
                if( object.ReferenceEquals( children[i], item ) ) return i;
            return -1;
        }

        public void Insert( int index, TemplateRow item )
        {
            children.Insert( index, item );
            
            Queue<TemplateRow> rowQueue = GetSubRows( index, 1 );

            TemplateRow templateRow;

            if( document.Changing )
            {
                while( rowQueue.Count > 0 )
                {
                    templateRow = rowQueue.Dequeue();
                    templateRow.Document = document;
                    templateRow.Parent = this;

                    RowAddedEvent( templateRow, EventArgs.Empty );
                }

                UpdateNumbering( index );
            }
            else
            {
                document.BeginChange();

                while( rowQueue.Count > 0 )
                {
                    templateRow = rowQueue.Dequeue();
                    templateRow.Document = document;
                    templateRow.Parent = this;

                    RowAddedEvent( templateRow, EventArgs.Empty );
                }

                UpdateNumbering( index );

                document.EndChange( false );
            }
        }
        
        public void InsertRange( int index, TemplateRow[] items )
        {
            children.Insert( index, items );
            
            Queue<TemplateRow> rowQueue = GetSubRows( index, items.Length );

            TemplateRow templateRow;

            if( document.Changing )
            {
                while( rowQueue.Count > 0 )
                {
                    templateRow = rowQueue.Dequeue();
                    templateRow.Document = document;

                    RowAddedEvent( templateRow, EventArgs.Empty );
                }

                UpdateNumbering( index );
            }
            else
            {
                document.BeginChange();

                while( rowQueue.Count > 0 )
                {
                    templateRow = rowQueue.Dequeue();
                    templateRow.Document = document;

                    RowAddedEvent( templateRow, EventArgs.Empty );
                }

                UpdateNumbering( index );

                document.EndChange( false );
            }
        }

        public bool Remove()
        {
            return parent.Remove( this );
        }

        public bool Remove( TemplateRow item )
        {
            int index = IndexOf( item );

            if( index == -1 ) return false;
            else
            {
                RemoveAt( index );

                return true;
            }
        }

        public void RemoveAt( int index )
        {
            Queue<TemplateRow> rowQueue = GetSubRows( index, 1 );

            children.Remove( index, 1 );

            if( document.Changing )
            {
                while( rowQueue.Count > 0 ) 
                    RowRemovedEvent( rowQueue.Dequeue(), EventArgs.Empty );

                UpdateNumbering( index );
            }
            else
            {
                document.BeginChange();

                while( rowQueue.Count > 0 ) 
                    RowRemovedEvent( rowQueue.Dequeue(), EventArgs.Empty );

                UpdateNumbering( index );

                document.EndChange( false );
            }
        }

        public void RemoveRange( int index, int length )
        {
            Queue<TemplateRow> rowQueue = GetSubRows( index, length );

            children.Remove( index, length );

            if( document.Changing )
            {
                while( rowQueue.Count > 0 ) 
                    RowRemovedEvent( rowQueue.Dequeue(), EventArgs.Empty );

                UpdateNumbering( index );
            }
            else
            {
                document.BeginChange();

                while( rowQueue.Count > 0 ) 
                    RowRemovedEvent( rowQueue.Dequeue(), EventArgs.Empty );

                UpdateNumbering( index );

                document.EndChange( false );
            }

        }

        protected void ObjectChangedEvent( object sender, EventArgs eventArgs )
        {
            if( objectChanging ) 
            {
                changeArgs = eventArgs;
                return;
            }

            if( ObjectChanged != null ) ObjectChanged( sender, eventArgs );
        }

        private void IndentChangedEvent( object sender, EventArgs eventArgs )
        {
            if( IndentChanged != null ) IndentChanged( sender, eventArgs );

            ObjectChangedEvent( sender, eventArgs );
        }

        private Queue<TemplateRow> GetSubRows( int index, int length )
        {
            Queue<TemplateRow> rowQueue = new Queue<TemplateRow>();

            for( int i = index; i < index + length; i++ )
            {
                foreach( TemplateRow templateRow in children[i] )
                {
                    rowQueue.Enqueue( templateRow ); 
                }
            }

            return rowQueue;
        }

        private void OnTextChanged( object sender, EventArgs eventArgs )
        {
            if( TextChanged != null ) TextChanged( sender, eventArgs );

            ObjectChangedEvent( sender, eventArgs );
        }

        private void RowAddedEvent( object sender, EventArgs eventArgs )
        {
            if( RowAdded != null ) RowAdded( sender, eventArgs );
        }

        private void RowRemovedEvent( object sender, EventArgs eventArgs )
        {
            if( RowRemoved != null ) RowRemoved( sender, eventArgs );
        }

        private void UpdateNumbering( int index )
        {
            INumbered previousRow = null;
            INumbered currentRow = null;

            if( index > 0 ) previousRow = children[index - 1] as INumbered;

            for( int i = index; i < children.Length; i++ )
            {
                currentRow = children[i] as INumbered;

                if( currentRow == null ) break;

                if( previousRow == null ) currentRow.Numbering.Reset();
                else
                {
                    currentRow.Numbering.Copy( previousRow.Numbering, currentRow.Numbering.Level );
                    currentRow.Numbering.Increase( 1 );
                }

                previousRow = currentRow;
                currentRow = null;
            }
        }

        IEnumerator IEnumerable.GetEnumerator()
        {
            throw new NotImplementedException();
        }
    }
}
