﻿using System;
using System.Collections.Generic;
using System.Text;
using Editor.Collections;
using Editor.Commands;
using Editor.Document.UseCase;
using Editor.Numbering;
using Editor.Text;
using Editor.View;
using Editor.Project;
using Editor.Commands.Document;
using System.Collections;

namespace Editor.Document
{
    public sealed class UseCaseDocument : IDocument, IEnumerable<TemplateRow>
    {
        private GapBuffer<TemplateRow> children;
        private IUndoableCommand lastCommand;
        private EventArgs changeArgs;
        private List<IUndoableCommand> commandHistory;
        private int currentCommnad;
        private List<IDocumentRow> documentRows;
        private UseCaseModel model;
        private string name;
        private bool objectChanging;
        private UseCaseProject project;
        private bool rowsChanged;
        private ISelection selection;

        public IDocumentRow this[int index]
        {
            get { return documentRows[index]; }
        }

        public bool Changing
        {
            get { return objectChanging; }
        }

        public int ItemCount
        {
            get { return children.Length; }
        }

        public IUndoableCommand LastCommand
        {
            get { return lastCommand; }
        }

        public UseCaseModel Model
        {
            get { return model; }
        }

        public string Name
        {
            get { return name; }
            set { name = value; }
        }

        public UseCaseProject Project
        {
            get { return project; }
        }

        public IUndoableCommand RedoCommand
        {
            get 
            { 
                if( currentCommnad == commandHistory.Count - 1 ) return null;
    
                return commandHistory[currentCommnad + 1];
            }
        }

        public int RowCount
        {
            get { return documentRows.Count; }
        }

        public ISelection Selection
        {
            get { return selection; }
        }

        public IUndoableCommand UndoCommand
        {
            get 
            { 
                if( currentCommnad < 0 ) return null;

                return commandHistory[currentCommnad];
            }
        }

        public event EventHandler RowAdded;
        public event EventHandler RowChanged;
        public event EventHandler RowRemoved;
        public event EventHandler ObjectChanged;
        public event EventHandler ObjectChanging;

        public UseCaseDocument( string name, UseCaseModel model )
        {
            children = new GapBuffer<TemplateRow>();
            commandHistory = new List<IUndoableCommand>();
            currentCommnad = -1;
            documentRows = new List<IDocumentRow>();
            lastCommand = null;
            this.model = model;
            this.name = name;
            project = model.ParentItem as UseCaseProject;
            selection = new DocumentSelection();

            InitializeStructure();
        }

        public void Add( TemplateRow item )
        {
            Insert( children.Length, item );
        }

        public void BeginChange()
        {
            if( !objectChanging )
            {
                objectChanging = true;
                changeArgs = null;

                if( ObjectChanging != null ) ObjectChanging( this, EventArgs.Empty );
            }
        }

        public void Clear()
        {
            Queue<TemplateRow> rowQueue = GetSubRows( 0, children.Length );

            children.SetBuffer( new TemplateRow[0] );

            if( objectChanging )
            {
                while( rowQueue.Count > 0 ) 
                    OnRowRemoved( rowQueue.Dequeue(), EventArgs.Empty );
            }
            else
            {
                BeginChange();

                while( rowQueue.Count > 0 ) 
                    OnRowRemoved( rowQueue.Dequeue(), EventArgs.Empty );

                EndChange( false );
            }
        }

        public void EndChange( bool discardEvent )
        {
            if( objectChanging )
            {
                objectChanging = false;

                if( !discardEvent && changeArgs != null ) ObjectChangedEvent( this, changeArgs ); 
            }
        }

        public bool ExecuteCommand( ICommand command )
        {
            IDocumentCommand documentCommand = command as IDocumentCommand;
            bool executed = false;

            if( documentCommand == null ) executed = command.Execute();
            else
            {
                if( !object.ReferenceEquals( this, documentCommand.Document ) )
                    throw new ArgumentException("Cannot execute foreign command", "command");

                if( !objectChanging )
                {
                    BeginChange();

                    executed = command.Execute();

                    EndChange( false );
                }
                else executed = command.Execute();
            }

            if( command is IUndoableCommand && executed )
            {
                this.TrimHistory();
                commandHistory.Add( command as IUndoableCommand );
                currentCommnad++;

                lastCommand = command as IUndoableCommand;
            }

            return executed;
        }

        public int FindSection( ESectionType sectionType )
        {
            DocumentSection section = null;

            for( int i = 0; i < documentRows.Count; i++ )
            {
                section = documentRows[i] as DocumentSection;

                if( section != null && section.Type == sectionType ) return i; 
            }

            return -1;
        }

        public Delegate[] GetReferences()
        {
            return ObjectChanged.GetInvocationList();
        }

        public TemplateRow GetTemplateRow( int index )
        {
            return children[index];
        }

        int IDocument.IndexOf( IDocumentRow documentRow )
        {
            return documentRows.IndexOf( documentRow );
        }

        public int IndexOf( TemplateRow item )
        {
            for( int i = 0; i < children.Length; i++ )
                if( object.ReferenceEquals( children[i], item ) ) return i;
            return -1;
        }

        public void Insert( int index, TemplateRow item )
        {
            children.Insert( index, item );
            
            Queue<TemplateRow> rowQueue = GetSubRows( index, 1 );

            TemplateRow templateRow;

            if( Changing )
            {
                while( rowQueue.Count > 0 )
                {
                    templateRow = rowQueue.Dequeue();
                    templateRow.Document = this;

                    OnRowAdded( templateRow, EventArgs.Empty );
                }

                UpdateNumbering( index );
            }
            else
            {
                BeginChange();

                while( rowQueue.Count > 0 )
                {
                    templateRow = rowQueue.Dequeue();
                    templateRow.Document = this;

                    OnRowAdded( templateRow, EventArgs.Empty );
                }

                UpdateNumbering( index );

                EndChange( false );
            }
        }

        public IEnumerator<TemplateRow> GetEnumerator()
        {
            for( int i = 0; i < children.Length; i++ )
            {
                IEnumerator<TemplateRow> enumerator = children[i].GetEnumerator();
                while( enumerator.MoveNext() )
                    yield return enumerator.Current;
            }
        }

        public void RecacheRows()
        {
            for( int i = 0; i < documentRows.Count; i++ )
            {
                UnregisterEvents( documentRows[i] as TemplateRow );
            }

            documentRows.Clear();

            IEnumerator<TemplateRow> enumerator = GetEnumerator();
                
            while( enumerator.MoveNext() )
            {
                RegisterEvents( enumerator.Current );

                documentRows.Add( enumerator.Current as IDocumentRow );
            }

            rowsChanged = false;
        }

        public void Redo()
        {
            if( currentCommnad == commandHistory.Count - 1 ) return;

            IUndoableCommand command = commandHistory[currentCommnad + 1];
            command.Execute();
            currentCommnad++;

            lastCommand = null;
        }

        public bool Remove( TemplateRow item )
        {
            int index = IndexOf( item );

            if( index == -1 ) return false;
            else
            {
                RemoveAt( index );
                return true;
            }
        }

        public void RemoveAt( int index )
        {
            Queue<TemplateRow> rowQueue = GetSubRows( index, 1 );

            children.Remove( index, 1 );

            if(  objectChanging )
            {
                while( rowQueue.Count > 0 ) 
                    OnRowRemoved( rowQueue.Dequeue(), EventArgs.Empty );

                UpdateNumbering( index );
            }
            else
            {
                BeginChange();

                while( rowQueue.Count > 0 ) 
                    OnRowRemoved( rowQueue.Dequeue(), EventArgs.Empty );

                UpdateNumbering( index );

                EndChange( false );
            }
        }

        public void Undo()
        {
            if( currentCommnad < 0 ) return;

            IUndoableCommand command = commandHistory[currentCommnad];
            command.Undo();
            currentCommnad--;

            lastCommand = null;
        }

        private void InitializeStructure()
        {
            BeginChange();

            DocumentSection section = new DocumentSection( "Actors", ESectionType.Actors );
            Add( section );
            section.Add( new Actors() );

            Add( new EmptyLine() );

            for( int i = 0; i < project.DescriptionItems.Count; i++ )
            {
                section = new DocumentSection( project.DescriptionItems[i].Name, ESectionType.Attribute );
                Add( section );
                section.Add( new Paragraph() );

                Add( new EmptyLine() );
            }

            section = new DocumentSection( "Main Flows", ESectionType.MainFlow );
            Add( section );

            Add( new EmptyLine() );

            section = new DocumentSection( "Alternative Flows", ESectionType.AlternativeFlow );
            Add( section );

            Add( new EmptyLine() );

            section = new DocumentSection( "Subflows", ESectionType.SubFlow );
            Add( section );

            EndChange(false);
        }

        private Queue<TemplateRow> GetSubRows( int index, int length )
        {
            Queue<TemplateRow> rowQueue = new Queue<TemplateRow>();

            for( int i = index; i < index + length; i++ )
            {
                foreach( TemplateRow templateRow in children[i] )
                {
                    rowQueue.Enqueue( templateRow ); 
                }
            }

            return rowQueue;
        }

        private void ObjectChangedEvent( object sender, EventArgs eventArgs )
        {
            if( objectChanging ) 
            {
                changeArgs = eventArgs;
                return;
            }

            if( rowsChanged ) RecacheRows();

            if( ObjectChanged != null ) ObjectChanged( sender, eventArgs );
        }

        private void OnRowAdded( object sender, EventArgs eventArgs )
        {
            rowsChanged = true;

            if( RowAdded != null ) RowAdded( sender, eventArgs );

            ObjectChangedEvent( sender, eventArgs );
        }

        private void OnRowChanged( object sender, EventArgs eventArgs )
        {
            if( RowChanged != null ) RowChanged( sender, eventArgs );

            ObjectChangedEvent( sender, eventArgs );
        }

        private void OnRowRemoved( object sender, EventArgs eventArgs )
        {
            rowsChanged = true;

            if( RowRemoved != null ) RowRemoved( sender, eventArgs );

            ObjectChangedEvent( sender, eventArgs );
        }

        private void RegisterEvents( TemplateRow templateRow )
        {
            templateRow.RowAdded += OnRowAdded;
            templateRow.ObjectChanged += OnRowChanged;
            templateRow.RowRemoved += OnRowRemoved;
        }

        private void TrimHistory()
        {
            if( commandHistory.Count == 0 ) return;
            if( currentCommnad == commandHistory.Count - 1 ) return;

            for( int i = commandHistory.Count - 1; i > currentCommnad; i-- )
                commandHistory.RemoveAt(i);
        }

        private void UnregisterEvents( TemplateRow templateRow )
        {
            templateRow.RowAdded -= OnRowAdded;
            templateRow.ObjectChanged -= OnRowChanged;
            templateRow.RowRemoved -= OnRowRemoved;
        }

        private void UpdateNumbering( int index )
        {
            INumbered previousRow = null;
            INumbered currentRow = null;

            if( index > 0 ) previousRow = children[index - 1] as INumbered;

            for( int i = index; i < children.Length; i++ )
            {
                currentRow = children[i] as INumbered;

                if( currentRow == null ) break;

                if( previousRow == null ) currentRow.Numbering.Reset();
                else
                {
                    currentRow.Numbering.Copy( previousRow.Numbering, currentRow.Numbering.Level );
                    currentRow.Numbering.Increase( 1 );
                }

                previousRow = currentRow;
            }
        }

        IEnumerator IEnumerable.GetEnumerator()
        {
            throw new NotImplementedException();
        }
    }
}
