﻿using System;
using System.Collections.Generic;
using System.Text;

namespace Editor.Numbering
{
    public sealed class DefaultNumbering : INumbering
    {
        private NumberingArgs changeArgs;
        private string label;
        private List<ShortNumber> numbers;
        private bool objectChanging;

        public bool Changing
        {
            get { return objectChanging; }
        }

        public string Label 
        {
            get { return label; }
            set { label = value; }
        }

        public int Level
        {
            get { return numbers.Count; }
        }

        public string LongestNumber
        {
            get
            {
                StringBuilder buffer = new StringBuilder();

                buffer.Append( label );
                for( int i = 0; i < numbers.Count - 1; i++ )
                {
                    buffer.Append( numbers[i].Longest );
                    buffer.Append('.');
                }
                buffer.Append( numbers[numbers.Count - 1].Longest);

                return buffer.ToString();
            }
        }

        internal List<ShortNumber> Numbers
        {
            get { return numbers; }
        }

        public event EventHandler ObjectChanged;
        public event EventHandler ObjectChanging;

        public DefaultNumbering() 
        {
            label = string.Empty;
            numbers = new List<ShortNumber>();
            
            numbers.Add( new ShortNumber() );
        }

        public DefaultNumbering( string label ) 
        {
            changeArgs = null;
            this.label = label;
            numbers = new List<ShortNumber>();
            objectChanging = false;
            
            numbers.Add( new ShortNumber() );
        }

        public void BeginChange()
        {
            if( !objectChanging )
            {
                objectChanging = true;
                changeArgs = null;

                if( ObjectChanging != null ) ObjectChanging( this, EventArgs.Empty );
            }
        }

        public INumbering Clone()
        {
            DefaultNumbering defaultNumbering = new DefaultNumbering();
            
            defaultNumbering.Label = label;
            defaultNumbering.Numbers.Clear();

            for( int i = 0; i < defaultNumbering.Numbers.Count; i++ )
            {
                defaultNumbering.Numbers.Add( defaultNumbering.Numbers[i].Clone() );
            }

            return defaultNumbering;
        }

        public void Copy( INumbering other, int level )
        {
            if( other == null ) throw new ArgumentNullException("other");

            DefaultNumbering defaultNumbering = other as DefaultNumbering;

            if( defaultNumbering == null )
                throw new ArgumentException("Argument not " + this.GetType().Name, "other");

            numbers.Clear();

            for( int i = 0; i < level; i++ )
            {
                numbers.Add( defaultNumbering.Numbers[i].Clone() );
            }
        }

        public bool Decrease( short difference )
        {
            bool executed = numbers[numbers.Count - 1].Decrease( difference );

            if( executed ) ObjectChangedEvent( this, new NumberingArgs( Convert.ToInt16( -difference ), 0 ) );
            
            return executed;
        }

        public bool DecreaseLevel()
        {
            if( numbers.Count > 1 )
            {
                numbers.RemoveAt( numbers.Count - 1 );

                ObjectChangedEvent( this, new NumberingArgs( 0, -1 ) );

                return true;
            }
            return false;
        }

        public void EndChange( bool discardEvent )
        {
            if( objectChanging )
            {
                objectChanging = false;

                if( !discardEvent && changeArgs != null ) ObjectChangedEvent( this, changeArgs ); 
            }
        }

        public Delegate[] GetReferences()
        {
            return ObjectChanged.GetInvocationList();
        }

        public bool Increase( short difference )
        {
            bool executed = numbers[numbers.Count - 1].Increase( difference );

            if( executed) ObjectChangedEvent( this, new NumberingArgs( difference, 0 ) );

            return executed;
        }

        public bool IncreaseLevel()
        {
            if( numbers.Count < 8 ) 
            {
                numbers.Add( new ShortNumber() );

                ObjectChangedEvent( this, new NumberingArgs( 0, 1 ) );

                return true;
            }
            return false;
        }

        public INumbering NextNumber()
        {
            INumbering number = this.Clone();

            if( number.Increase( 1 ) ) return number;
            else return null;

        }

        public INumbering PreviousNumber()
        {
            INumbering number = this.Clone();

            if( number.Decrease( 1 ) ) return number;
            else return null;
        }

        public override string ToString()
        {
            StringBuilder stringBuilder = new StringBuilder();

            stringBuilder.Append( label );
            for( int i = 0; i < numbers.Count - 1; i++ )
            {
                stringBuilder.Append( numbers[i].Current );
                stringBuilder.Append('.');
            }
            stringBuilder.Append( numbers[numbers.Count - 1].Current );

            return stringBuilder.ToString();
        }

        public int CompareTo( INumbering other )
        {
            if( other == null ) throw new ArgumentNullException("other");

            DefaultNumbering defaultNumbering = other as DefaultNumbering;

            if( defaultNumbering == null )
                return string.Compare( this.ToString(), other.ToString() );

            DefaultNumbering otherNumbering = other as DefaultNumbering;
            int comparison = 0;

            if( numbers.Count < defaultNumbering.Numbers.Count )
            {
                for( int i = 0; i < numbers.Count; i++ )
                {
                    comparison = numbers[i].CompareTo( defaultNumbering.Numbers[i] );
                    if( comparison != 0 ) return comparison;
                }
                
                comparison = -1;
            }
            
            if( numbers.Count > defaultNumbering.Numbers.Count )
            {
                for( int i = 0; i < defaultNumbering.Numbers.Count; i++ )
                {
                    comparison = numbers[i].CompareTo( defaultNumbering.Numbers[i] );
                    if( comparison != 0 ) return comparison;
                }

                comparison = 1;
            }
            
            return comparison;
        }

        public void Reset()
        {
            numbers.Clear();
            numbers.Add( new ShortNumber() );
        }

        private void ObjectChangedEvent( object sender, NumberingArgs eventArgs )
        {
            if( objectChanging ) 
            {
                if( changeArgs == null ) changeArgs = eventArgs;
                else
                {
                    changeArgs = new NumberingArgs( (short)(eventArgs.Level + changeArgs.Level),
                        (short)(eventArgs.Number + changeArgs.Number) );
                }

                return;
            }

            if( ObjectChanged != null ) ObjectChanged( sender, eventArgs );
        }
    }
}
