﻿using System;
using System.Collections.Generic;
using System.Text;
using System.Windows.Forms;

namespace Editor.Project
{
    public sealed class UseCaseModel : TreeNode, IProjectItem
    {
        private EventArgs changeArgs;
        private bool objectChanging;
        private UseCaseProject project;
        private bool saved;

        public IProjectItem this[int index]
        {
            get { return Nodes[index] as IProjectItem; }
        }

        public bool Composite
        {
            get { return true; }
        }

        public bool Changing
        {
            get { return objectChanging; }
        }

        public int ItemCount
        {
            get { return Nodes.Count; }
        }

        public string ItemName
        {
            get { return Text; }
            set 
            { 
                Text = value;

                ObjectChangedEvent( this, EventArgs.Empty );
            }
        }

        public IProjectItem ParentItem
        {
            get { return project as IProjectItem; }
        }

        public bool Saved
        {
            get { return saved; }
            set
            {
                saved &= value;
            }
        }

        public event EventHandler ObjectChanged;
        public event EventHandler ObjectChanging;

        public UseCaseModel( string name, UseCaseProject project )
        {
            ImageIndex = 0;
            this.project = project;
            saved = false;
            SelectedImageIndex = 0;
            Text = name;
        }

        public void Add( IProjectItem item )
        {
            Insert( Nodes.Count, item );
        }

        public void BeginChange()
        {
            if( !objectChanging )
            {
                objectChanging = true;
                changeArgs = null;

                if( ObjectChanging != null ) ObjectChanging( this, EventArgs.Empty );
            }
        }

        public void Clear()
        {
            Nodes.Clear();

            ObjectChangedEvent( this, EventArgs.Empty );
        }

        public IProjectItem CreateSubItem( Arguments argumets )
        {
            if( argumets is ActorArguments )
            {
                Actor actor = new Actor( (argumets as ActorArguments).Name, this );
                
                Add( actor );

                return actor;
            }

            if( argumets is UseCaseArguments )
            {
                UseCase useCase = new UseCase( (argumets as UseCaseArguments).Name, this );
                
                Add( useCase );

                return useCase;
            }

            return null;
        }

        public void EndChange( bool discardEvent )
        {
            if( objectChanging )
            {
                objectChanging = false;

                if( !discardEvent && changeArgs != null ) ObjectChangedEvent( this, changeArgs ); 
            }
        }

        public Delegate[] GetReferences()
        {
            return ObjectChanged.GetInvocationList();
        }

        public int IndexOf( IProjectItem item )
        {
            TreeNode itemNode = item as TreeNode;

            if( itemNode == null ) return -1;

            return Nodes.IndexOf( itemNode );
        }

        public int IndexOf( IProjectItem item, int startIndex )
        {
            for( int i = startIndex; i < Nodes.Count; i++ )
            {
                if( object.ReferenceEquals( item, Nodes[i] ) ) return i; 
            }

            return -1;
        }

        public void Insert( int index, IProjectItem item )
        {
            TreeNode itemNode = item as TreeNode;

            if( itemNode == null ) return;

            if( item is UseCase )
            {
                item.ObjectChanged += OnProjectItemChanged;

                for( int i = 0; i < Nodes.Count; i++ )
                {
                    UseCase useCase = Nodes[i] as UseCase;

                    if( useCase != null )
                    {
                        if( string.Compare( item.ItemName, useCase.ItemName ) < 0 )
                        {
                            Nodes.Insert( i, itemNode );
                            return;
                        }
                    }
                    else
                    {
                        Nodes.Insert( 0, itemNode );
                        return;
                    }
                }

                Nodes.Insert( Nodes.Count, itemNode );

                ObjectChangedEvent( this, EventArgs.Empty );

                return;
            }

            if( item is Actor )
            {
                item.ObjectChanged += OnProjectItemChanged;

                for( int i = 0; i < Nodes.Count; i++ )
                {
                    Actor actor = Nodes[i] as Actor;

                    if( actor != null )
                    {
                        if( string.Compare( item.ItemName, actor.ItemName ) < 0 )
                        {
                            Nodes.Insert( i, itemNode );
                            return;
                        }
                    }
                }

                Nodes.Insert( Nodes.Count, itemNode );

                ObjectChangedEvent( this, EventArgs.Empty );

                return;
            }
        }

        public void InsertRange( int index, IProjectItem[] items )
        {
            throw new NotImplementedException();
        }

        public bool Load( Arguments arguments )
        {
            throw new NotImplementedException();
        }

        public bool Remove( IProjectItem item )
        {
            TreeNode itemNode = item as TreeNode;

            if( itemNode == null ) return false;

            item.ObjectChanged -= OnProjectItemChanged;

            Nodes.Remove( itemNode );

            ObjectChangedEvent( this, EventArgs.Empty );

            return true;
        }

        public void RemoveAt( int index )
        {
            IProjectItem item = Nodes[index] as IProjectItem;

            if( item != null ) item.ObjectChanged -= OnProjectItemChanged;

            Nodes.RemoveAt( index );

            ObjectChangedEvent( this, EventArgs.Empty );
        }

        public void RemoveRange( int index, int length )
        {
            throw new NotImplementedException();
        }
        
        public bool Save( Arguments arguments )
        {
            throw new NotImplementedException();
        }

        private void ObjectChangedEvent( object sender, EventArgs eventArgs )
        {
            if( objectChanging ) 
            {
                changeArgs = eventArgs;
                return;
            }

            if( ObjectChanged != null ) ObjectChanged( sender, eventArgs );
        }

        private void OnProjectItemChanged( object sender, EventArgs eventArgs )
        {
            ObjectChangedEvent( sender, EventArgs.Empty );
        }
    }
}
