﻿using System;
using System.Collections.Generic;
using System.Text;
using Editor.Collections;
using System.Drawing;

namespace Editor.Text
{
    public sealed class GlyphText : IText
    {
        private EventArgs changeArgs;
        private string characters;
        private int characterLength;
        private GapBuffer<Glyph> glyphBuffer;
        private bool objectChanging;

        public Glyph this[int index]
        {
            get { return glyphBuffer[index]; }
        }

        public bool Changing
        {
            get { return objectChanging; }
        }

        public int CharacterLength
        {
            get { return characterLength; }
        }

        public bool Empty
        {
            get { return glyphBuffer.Length == 0; }
        }

        public int VirtualLength
        {
            get { return glyphBuffer.Length; }
        }

        internal GapBuffer<Glyph> Glyphs
        {
            get { return glyphBuffer; }
        }

        public event EventHandler ObjectChanged;
        public event EventHandler ObjectChanging;

        public GlyphText()
        {
            characters = string.Empty;
            characterLength = 0;
            glyphBuffer = new GapBuffer<Glyph>();
        }

        public GlyphText( Glyph[] glyphArray )
        {
            characters = string.Empty;
            characterLength = 0;

            glyphBuffer = new GapBuffer<Glyph>( glyphArray );

            IChangeHandler changeable = null;

            for( int i = 0; i < glyphBuffer.Length; i++ )
            {
                characterLength += glyphBuffer[i].CharacterLength;

                changeable = glyphBuffer[i] as IChangeHandler;
                
                if( changeable != null )
                    changeable.ObjectChanged += OnGlyphChanged;
            }
        }

        public void BeginChange()
        {
            if( !objectChanging )
            {
                objectChanging = true;
                changeArgs = null;

                if( ObjectChanging != null ) ObjectChanging( this, EventArgs.Empty );
            }
        }

        public IText ChangeStyle( int index, int length, short style )
        {
            GlyphText text = new GlyphText();
            text.Glyphs.SetBuffer( glyphBuffer.GetRange( index, length ) );

            for( int i = index; i < length; i++ )
            {
                glyphBuffer[i] = glyphBuffer[i].Clone();
                glyphBuffer[i].Style = style;
            }

            ObjectChangedEvent( this, EventArgs.Empty );

            return text;
        }

        public IText Copy( int index, int length )
        {
            Glyph[] glyphArray = glyphBuffer.GetRange( index, length );

            for( int i = 0; i < glyphArray.Length; i++ )
                glyphArray[i] = glyphArray[i].Clone();

            GlyphText text = new GlyphText( glyphArray );

            return text;
        }

        public IText CreateInstance()
        {
            return new GlyphText();
        }

        public Delegate[] GetReferences()
        {
            return ObjectChanged.GetInvocationList();
        }

        public short GetStyle( int index )
        {
            return glyphBuffer[index].Style;
        }

        public void EndChange( bool discardEvent )
        {
            if( objectChanging )
            {
                objectChanging = false;

                if( !discardEvent && changeArgs != null ) ObjectChangedEvent( this, changeArgs ); 
            }
        }

        public short GetInsertStyle( int index )
        {
            short style = (short)EDefaultStyles.Normal;
  
            if( index == 0 )
            {
                if( glyphBuffer.Length > 0 && glyphBuffer[0].StyleCopying ) 
                    style = glyphBuffer[0].Style;
            }
            else
            {
                if( glyphBuffer[index - 1].StyleCopying ) style = glyphBuffer[index - 1].Style;
                else if( index != glyphBuffer.Length && glyphBuffer[index].StyleCopying ) 
                    style = glyphBuffer[index].Style;
            }

            return style;
        }

        public Glyph[] GetRange( int index, int length )
        {
            return glyphBuffer.GetRange( index, length );
        }

        public int GetVirtualIndex( int charIndex, out int startIndex, out int endIndex )
        {
            int charLength = 0;

            for( int i = 0; i < glyphBuffer.Length; i++ )
            {
                if( charLength + glyphBuffer[i].CharacterLength > charIndex )
                {
                    if( glyphBuffer[i].CharacterLength > 1 )
                    {
                        startIndex = charLength;
                        endIndex = charLength + glyphBuffer[i].CharacterLength - 1;
                        return i;
                    }
                    else
                    {
                        startIndex = charLength;
                        endIndex = charLength;
                        return i;
                    }
                }

                charLength += glyphBuffer[i].CharacterLength;
            }

            if( charIndex == charLength )
            {
                startIndex = charLength;
                endIndex = charLength;
                return glyphBuffer.Length;
            }

            startIndex = -1;
            endIndex = -1;
            return -1;
        }

        public IText InsertCharacter( int index, char character )
        {
            GlyphText text = new GlyphText( new Glyph[]{ new CharacterGlyph( character, 
                GetInsertStyle( index ) ) } );

            return InsertText( index, text );
        }

        public IText InsertGlyph( int index, Glyph glyph )
        {
            GlyphText text = new GlyphText( new Glyph[]{ glyph } );

            return InsertText( index, text );
        }

        public IText InsertString( int index, string characters )
        {
            short style = GetInsertStyle( index );
            Glyph[] glyphArray = new Glyph[characters.Length];

            for( int i = 0; i < characters.Length; i++ )
                glyphArray[i] = new CharacterGlyph( characters[i] );

            GlyphText text = new GlyphText( glyphArray );

            return InsertText( index, text );
        }

        public IText InsertText( int index, IText text )
        {
            IChangeHandler changeable = null;
            Glyph[] glyphArray = text.GetRange( 0, text.VirtualLength );

            glyphBuffer.Insert( index, glyphArray );

            for( int i = 0; i < glyphArray.Length; i++ )
            {
                characterLength += glyphArray[i].CharacterLength;

                changeable = glyphArray[i] as IChangeHandler;
                
                if( changeable != null )
                    changeable.ObjectChanged += OnGlyphChanged;
            }

            characters = string.Empty;

            ObjectChangedEvent( this, EventArgs.Empty );

            return text;
        }

        public IText Remove( int index, int length )
        {
            IChangeHandler changeable = null;
            Glyph[] glyphArray = glyphBuffer.GetRange( index, length );
            GlyphText removedText = new GlyphText( glyphArray );

            glyphBuffer.Remove( index, length );
            characterLength -= removedText.CharacterLength;

             for( int i = 0; i < glyphArray.Length; i++ )
            {
                changeable = glyphArray[i] as IChangeHandler;
                
                if( changeable != null )
                    changeable.ObjectChanged -= OnGlyphChanged;
            }

            characters = string.Empty;
            
            ObjectChangedEvent( this, EventArgs.Empty );

            return removedText;
        }

        public override string ToString()
        {
            if( glyphBuffer.Length == 0 ) return string.Empty;
            else
            {
                if( characters == string.Empty )
                {
                    StringBuilder buffer = new StringBuilder();

                    for( int i = 0; i < glyphBuffer.Length; i++ )
                        buffer.Append( glyphBuffer[i].ToString() );

                    characters = buffer.ToString();
                }
                
                return characters;
            }
        }

        private void ObjectChangedEvent( object sender, EventArgs eventArgs )
        {
            if( objectChanging ) 
            {
                changeArgs = eventArgs;
                return;
            }

            if( ObjectChanged != null ) ObjectChanged( sender, eventArgs );
        }

        private void OnGlyphChanged( object sender, EventArgs eventArgs )
        {
            ObjectChangedEvent( this, EventArgs.Empty );
        }
    }
}