﻿using System;
using System.Collections.Generic;
using System.Text;
using System.Drawing;
using Editor.Document;
using Editor.Text;
using Editor.Numbering;

namespace Editor.View
{
    public sealed class ViewRow : IViewRow
    {
        private IDocumentRow documentRow;
        private int height;
        private int textIndent;
        private int offset;
        private List<IViewLine> segmentLines;
        private IView view;

        public IDocumentRow DocumentRow
        {
            get { return documentRow; }
        }

        public int TextIndent
        {
            get { return textIndent; }
        }

        public int Height
        {
            get { return height; }
        }

        public int Offset 
        {
            get { return offset; }
            set { offset = value; }
        }

        public IView View
        {
            get { return view; }
        }

        public ViewRow( IView view, IDocumentRow documentRow )
        {
            this.documentRow = documentRow;
            textIndent = 0;
            segmentLines = new List<IViewLine>();
            this.view = view;

            segmentLines.Add( CreateLine() );
            height = segmentLines[0].Height;

            documentRow.IndentChanged += OnIndentChanged;
            documentRow.TextChanged += OnTextChanged;
            if( documentRow is INumbered ) (documentRow as INumbered).NumberingChanged += OnNumberingChanged;

            Prepare();
        }

        public void Draw()
        {
            int rowIndex = view.Document.IndexOf( documentRow );
            SSelection selection = view.Document.Selection.GetOrdered();
            int lineOffset = offset;

            for( int i = 0; i < segmentLines.Count; i++ )
            {
                if( ( selection.StartRow == rowIndex ) && ( selection.EndRow == rowIndex) )
                    segmentLines[i].Draw( lineOffset, selection.StartIndex, 
                        selection.EndIndex - 1 );
                else if( selection.StartRow == rowIndex )
                    segmentLines[i].Draw( lineOffset, selection.StartIndex );
                else if( selection.EndRow == rowIndex )
                    segmentLines[i].Draw( lineOffset, 0, selection.EndIndex - 1 );
                else
                {
                    if( view.Document.Selection.ContainsRow( rowIndex ) ) 
                        segmentLines[i].Draw( lineOffset, true );
                    else segmentLines[i].Draw( lineOffset, false );
                }

                lineOffset += segmentLines[i].Height;              
            }
        }

        public int GetSelectionIndex( int x, int y )
        {
            IViewLine line = null;
            int lineOffset = 0;

            for( int i1 = 0; i1 < segmentLines.Count; i1++ )
            {
                line = segmentLines[i1];
                lineOffset += line.Height;

                if( y - offset < lineOffset ) 
                {
                    int segmentOffset = view.Margins + textIndent;

                    if( x < segmentOffset )
                    {
                        line = segmentLines[i1];

                        for( int i2 = 0; i2 < line.ItemCount; i2++ )
                            if( line[i2].StartIndex >= 0 ) return line[i2].StartIndex;

                        return -1;
                    }
                    else
                    {
                        int endIndex = 0;
                        int selectionIndex = 0;

                        for( int i2 = 0; i2 < line.ItemCount; i2++ )
                        {
                            selectionIndex = line[i2].GetSelectionIndex( x );

                            if( selectionIndex == -1 )
                            {
                                if( line[i2].EndIndex >= 0 ) endIndex = line[i2].EndIndex + 1;
                            }
                            else return selectionIndex;
                        }

                        return endIndex;
                    }
                }
            }

            return 0;
        }

        private void Insert( int index, IViewLine item )
        {
            segmentLines.Insert( index, item );
        }

        public void PlaceCaret( int index, int y, int correction )
        {
            IViewLine line = null;
            ISegment segment = null;
            int lineOffset = offset;

            if( y < lineOffset ) view.Caret.Place( view.Margins + textIndent, lineOffset, 
                segmentLines[0].Height, true );

            for( int i1 = 0; i1 < segmentLines.Count; i1++ )
            {
                line = segmentLines[i1];

                for( int i2 = 0; i2 < line.ItemCount; i2++ )
                {
                    if( line[i2].StartIndex == -1 ) continue;

                    segment = line[i2];

                    if( segment.StartIndex <= index && index <= segment.EndIndex ) 
                    {
                        Size size = Size.Empty;
                        
                        size = segment.Measure( segment.StartIndex, index - segment.StartIndex );
     
                        view.Caret.Place( segment.Indent + size.Width + correction, lineOffset, 
                            line.Height, true );
                        return;
                    }
                }

                lineOffset += line.Height;

                if( y < lineOffset ) break;
            }

            if( segment == null )
                view.Caret.Place( view.Margins + textIndent, lineOffset - line.Height, line.Height, true );
            else
            {
                view.Caret.Place( segment.Indent + segment.Width + correction, lineOffset - line.Height, 
                    line.Height, true );
            }
        }

        public void Prepare()
        {
            if( documentRow is INumbered )
            {
                INumbered numberedItem = documentRow as INumbered;

                IStyle style = view.StyleTable[numberedItem.NumberingStyle];
                Size size = view.MeasureString( numberedItem.EqualNumberIndent ? numberedItem.LongestNumber :
                    numberedItem.CurrentNumber, style.Font );
                int numberingLevel = numberedItem.Numbering.Level == 0 ? 0 : numberedItem.Numbering.Level - 1;


                textIndent = documentRow.LeftIndent + ( numberingLevel * view.RowLevelIndent ) + 
                    size.Width + 5;

                FormatText( documentRow.Text );

                segmentLines[0].Insert( 0, new NumberSegment( view, numberedItem.CurrentNumber, style, documentRow.LeftIndent + 
                    ( numberingLevel * view.RowLevelIndent ) + view.Margins, size.Width ) );

                Update();
            }
            else
            {
                textIndent = documentRow.LeftIndent;

                FormatText( documentRow.Text );
            }   
        }

        private void Add( IViewLine item )
        {
            segmentLines.Add( item );
        }

        private void Clear()
        {
            segmentLines.Clear();
            segmentLines.Add( CreateLine() );

            height = segmentLines[0].Height;
        }

        private IViewLine CreateLine()
        {
            return new ViewLine( this );
        }


        private void FormatText( IText text )
        {
            Clear();

            int charCount = 0;
            IViewLine currentLine = segmentLines[segmentLines.Count - 1];
            int currentIndent = view.Margins + textIndent;
            int length = view.Size.Width - textIndent;
            TextIterator textIterator = new TextIterator( text );
            int width = 0;

            if( !textIterator.NextChar() ) 
            {
                Update();
                return;
            }

            do
            {
                char character = textIterator.Character;
                width = MeasureString( textIterator, out charCount );

                if( char.IsWhiteSpace( character ) )
                {
                    if( view.Margins + view.Size.Width - currentIndent > width )
                        currentLine = PlaceCharacters( currentLine, currentIndent, textIterator,
                            charCount, out currentIndent );
                    else
                        currentLine = PlaceCharacters( currentLine, currentIndent, view.Margins + textIndent,
                            view.Margins + view.Size.Width, textIterator, charCount, out currentIndent);            
                }
                else
                {
                    if( view.Margins + view.Size.Width - currentIndent > width )
                        currentLine = PlaceCharacters( currentLine, currentIndent,
                            textIterator, charCount, out currentIndent );
                    else
                    {
                        if( width < length )
                        {
                            if( !object.ReferenceEquals( currentLine, segmentLines[0] ) ) Add( currentLine );
                            currentLine = CreateLine();

                            currentIndent = view.Margins + textIndent;

                            currentLine = PlaceCharacters( currentLine, currentIndent,
                                textIterator, charCount, out currentIndent );                  
                        }
                        else currentLine = PlaceCharacters( currentLine, currentIndent, view.Margins + textIndent, 
                            view.Margins + view.Size.Width, textIterator, charCount, out currentIndent );  
                    }               
                }
            } while ( textIterator.CanIterate() );

            if( currentLine.ItemCount > 0 && !object.ReferenceEquals( currentLine, segmentLines[0] ) )
                Add( currentLine );

            Update();
        }

        private int MeasureString( TextIterator position, out int charCount )
        {
            StringBuilder buffer = new StringBuilder();
            char character = position.Character;
            short lastStyle = position.Glyph.Style;
            Size size = Size.Empty;
            TextIterator textIterator = position.Clone();
            bool whiteSpace = char.IsWhiteSpace( position.Character );
            int width = 0;

            charCount = 0;

            do
            {
                character = textIterator.Character;

                if( char.IsWhiteSpace( character ) != whiteSpace ) break;

                if( position.Glyph.Style != lastStyle )
                {
                    size = view.MeasureString( buffer.ToString(),
                        view.StyleTable[lastStyle].Font );
                    width += size.Width;

                    charCount += buffer.Length;
                    buffer.Remove( 0, buffer.Length );
                }

                lastStyle = position.Glyph.Style;
                buffer.Append( character );
            } while( textIterator.NextChar() );

            if( buffer.Length > 0 )
            {
                size = view.MeasureString( buffer.ToString(),
                    view.StyleTable[lastStyle].Font );
                width += size.Width;

                charCount += buffer.Length;
            }

            return width;
        }

        private void OnIndentChanged(object sender, EventArgs eventArgs)
        {
            Prepare();
        }

        private void OnNumberingChanged(object sender, EventArgs eventArgs)
        {
            Prepare();
        }

        private void OnTextChanged(object sender, EventArgs eventArgs)
        {
            Prepare();
        }

        private IViewLine PlaceCharacters( IViewLine line, int indent, TextIterator position, int charCount, out int currentIndent )
        {
            StringBuilder buffer = new StringBuilder();
            char character = position.Character;
            int charIndex = position.CharIndex;
            short lastStyle = position.Glyph.Style;
            Rectangle rectangle = Rectangle.Empty;
            Size size = Size.Empty;
            TextIterator textIterator = position;
            IStyle style = null;

            currentIndent = indent;

            for( int i = 0; i < charCount; i++  )
            {
                character = textIterator.Character;
                
                if( position.Glyph.Style != lastStyle )
                {
                    style = view.StyleTable[lastStyle];
                    size = view.MeasureString( buffer.ToString(), style.Font );

                    line.Add( new CharacterSegment( view, buffer.ToString(), style,
                        charIndex, currentIndent, size.Width ) );

                    charIndex = textIterator.CharIndex;
                    currentIndent += size.Width; 
                    buffer.Remove( 0, buffer.Length );
                }

                buffer.Append( character );
                lastStyle = textIterator.Glyph.Style;
                textIterator.NextChar();
            }

            if( buffer.Length > 0 )
            {
                style = view.StyleTable[lastStyle];
                size = view.MeasureString( buffer.ToString(), style.Font );

                line.Add( new CharacterSegment( view, buffer.ToString(), style,
                    charIndex, currentIndent, size.Width ) );

                currentIndent += size.Width; 
            }

            return line;
        }

        private IViewLine PlaceCharacters( IViewLine line, int indent, 
            int minIndent, int maxIndent, TextIterator position, int charCount, out int currentIndent )
        {
            StringBuilder buffer = new StringBuilder();
            char character = position.Character;
            int charIndex = position.CharIndex;
            IViewLine currentLine = line;
            short lastStyle = position.Glyph.Style;
            Rectangle rectangle = Rectangle.Empty;
            Size size = Size.Empty;
            TextIterator textIterator = position;
            IStyle style = null;

            currentIndent = indent;

            for( int i = 0; i < charCount; i++  )
            {
                character = textIterator.Character;
                
                if( position.Glyph.Style != lastStyle )
                {
                    style = view.StyleTable[lastStyle];
                    size = view.MeasureString( buffer.ToString(), style.Font );

                    currentLine.Add( new CharacterSegment( view, buffer.ToString(), style,
                        charIndex, currentIndent, size.Width ) );

                    charIndex = textIterator.CharIndex;
                    currentIndent += size.Width; 
                    buffer.Remove( 0, buffer.Length );
                }

                buffer.Append( character );
                lastStyle = position.Glyph.Style;

                style = view.StyleTable[lastStyle];
                size = view.MeasureString( buffer.ToString(), style.Font );

                if( currentIndent + size.Width > maxIndent )
                {
                    buffer.Remove( buffer.Length - 1, 1 );

                    size = view.MeasureString( buffer.ToString(), style.Font );

                    currentLine.Add( new CharacterSegment( view, buffer.ToString(), style,
                        charIndex, currentIndent, size.Width ) );

                    if( !object.ReferenceEquals( currentLine, segmentLines[0] ) ) Add( currentLine );
                    currentLine = CreateLine();

                    charIndex = textIterator.CharIndex;
                    currentIndent += size.Width; 
                    buffer.Remove( 0, buffer.Length );
                    buffer.Append( character );
                    currentIndent = minIndent;
                }

                textIterator.NextChar();
            }

            if( buffer.Length > 0 )
            {
                style = view.StyleTable[lastStyle];
                size = view.MeasureString( buffer.ToString(), style.Font );

                currentLine.Add( new CharacterSegment( view, buffer.ToString(), style,
                    charIndex, currentIndent, size.Width ) );

                currentIndent += size.Width;
            }

            return currentLine;
        }

        private void RemoveAt( int index )
        {
            segmentLines.RemoveAt( index );
        }

        private void Update()
        {
            height = 0;

            for( int i = 0; i < segmentLines.Count; i++ )
                height += segmentLines[i].Height;
        }
    }
}
